module GraphQL
  module Streaming
    # A subscriber for a channel-query combo.
    #
    # @example Subscribe to a query in an ActionCable action, re-evaluating it when things change
    #   # Initialize `context` ahead of time so it can be closed over in the subscription block
    #   context = {}
    #
    #   context[:subscriber] = GraphQLSubscriber.new(self, query_id) do
    #     # Tell the schema how to re-fetch results:
    #     context[:current_user].reload
    #     MySchema.execute(query_string, context: context, variables: variables)
    #   end
    #
    #   # Run the query
    #   MySchema.execute(query_string, context: context, variables: variables)
    #
    #   # detect whether any subscriptions were added
    #   context[:subscriber].subscribed?
    #
    class ActionCableSubscriber
      CHANNEL_PREFIX = "graphql_subscription_"

      # @param [ActionCable::Channel::Base] The channel to push updates to
      # @param [Object] The query who the updates belong to (probably provided by the client)
      # @yield Reruns the query
      def initialize(channel, query_id, &query_exec)
        @channel = channel
        @query_id = query_id
        @query_exec = query_exec
        @subscribed = false
        @own_streams = []
      end

      # Trigger an event with arguments
      #
      # @example Trigger post_changed
      #   # First, subscribe with "{ post_changed(id: 1) { title } }"
      #   # Then, trigger the event:
      #   GraphQL::Streaming::ActionCableSubscriber.trigger(:post_changed, {id: 1})
      #
      # @param [Symbol] The subscription name to trigger
      # @param [Hash] Arguments to send with the subscription
      def self.trigger(subscription_handle, trigger_options = {})
        ActionCable.server.broadcast("#{CHANNEL_PREFIX}#{subscription_handle}", trigger_options)
      end

      # Subscribe to event named `subscription_handle`, but only
      # when called with arguments `arguments`
      # @param [String] the event name to subscribe to
      # @param [Hash] the arguments to subscribe to
      def register(subscription_handle, arguments)
        @subscribed = true
        handle = "#{CHANNEL_PREFIX}#{subscription_handle}"
        original_args = stringify_hash(arguments)

        @own_streams << @channel.stream_from(handle) do |trigger_json|
          trigger_args = JSON.parse(trigger_json)
          if original_args == trigger_args
            reevaluate_query
          end
        end
      end

      # @return [Boolean] True if this subscriber registered any subscriptions
      def subscribed?
        @subscribed
      end

      # Tell this subscriber to stop sending patches
      # @return [void]
      def close
        @channel.stop_specific_streams(@own_streams)
        @own_streams.clear
        @subscribed = false
        nil
      end

      private

      # Re-evaluate the given block
      # and send the result as a patch to the
      # root of the query result
      def reevaluate_query
        payload = {
          patch: {
            path: [],
            value:  @query_exec.call,
          },
          query_id: @query_id,
        }
        @channel.send_graphql_payload(payload)
      end

      def stringify_hash(value)
        case value
        when Hash
          value.inject({}) { |memo, (k, v)| memo[k.to_s] = stringify_hash(v); memo }
        when Array
          value.map { |v| stringify_hash(v) }
        else
          value
        end
      end
    end
  end
end
